@@ -25,89 +25,130 @@
 #include "internal.h"
 #include "memory.h"
 #include "syntax.h"
+#include "transform.h"
 
 #include <fnmatch.h>
 #include <argz.h>
 #include <string.h>
+#include <stdarg.h>
 
-/* We always have a toplevel entry for P_ROOT */
-#define P_ROOT   "augeas"
+/* Some popular labels that we use in /augeas */
+static const char *const s_augeas = "augeas";
+static const char *const s_files  = "files";
+static const char *const s_load   = "load";
+static const char *const s_pathx  = "pathx";
+static const char *const s_error  = "error";
+static const char *const s_pos    = "pos";
 
 #define TREE_HIDDEN(tree) ((tree)->label == NULL)
 
-static const char *pretty_label(const struct tree *tree) {
-    if (tree == NULL)
-        return "(no_tree)";
-    else if (tree->label == NULL)
-        return "(none)";
-    else
-        return tree->label;
-}
+static const char *const static_nodes[][2] = {
+    { AUGEAS_META_TREE "/version", PACKAGE_VERSION },
+    { AUGEAS_META_TREE "/version/save/mode[1]", AUG_SAVE_BACKUP_TEXT },
+    { AUGEAS_META_TREE "/version/save/mode[2]", AUG_SAVE_NEWFILE_TEXT },
+    { AUGEAS_META_TREE "/version/save/mode[3]", AUG_SAVE_NOOP_TEXT },
+    { AUGEAS_META_TREE "/version/save/mode[4]", AUG_SAVE_OVERWRITE_TEXT },
+    { AUGEAS_META_TREE "/version/defvar", NULL }
+};
 
-static char *path_expand(struct tree *tree, const char *ppath) {
-    struct tree *siblings = tree->parent->children;
-
-    char *path;
-    const char *label;
-    int cnt = 0, ind = 0, r;
-
-    list_for_each(t, siblings) {
-        if (streqv(t->label, tree->label)) {
-            cnt += 1;
-            if (t == tree)
-                ind = cnt;
-        }
-    }
-
-    if (ppath == NULL)
-        ppath = "";
-
-    label = pretty_label(tree);
-    if (cnt > 1) {
-        r = asprintf(&path, "%s/%s[%d]", ppath, label, ind);
-    } else {
-        r = asprintf(&path, "%s/%s", ppath, label);
-    }
-    if (r == -1)
-        return NULL;
-    return path;
-}
-
-static char *format_path(struct tree *tree) {
-    int depth, i;
-    struct tree *t, **anc;
-    char *path = NULL;
-
-    for (t = tree, depth = 1; ! ROOT_P(t); depth++, t = t->parent);
-    if (ALLOC_N(anc, depth) < 0)
-        return NULL;
-
-    for (t = tree, i = depth - 1; i >= 0; i--, t = t->parent)
-        anc[i] = t;
-
-    for (i = 0; i < depth; i++) {
-        char *p = path_expand(anc[i], path);
-        free(path);
-        path = p;
-    }
-    return path;
-}
-
-/* Propagate dirty flags towards the root */
-static int tree_propagate_dirty(struct tree *tree) {
-    if (tree->dirty)
-        return 1;
-    list_for_each(c, tree->children) {
-        tree->dirty |= tree_propagate_dirty(c);
-    }
-    return tree->dirty;
+static void tree_mark_dirty(struct tree *tree) {
+    do {
+        tree->dirty = 1;
+        tree = tree->parent;
+    } while (tree != tree->parent && !tree->dirty);
 }
 
 /* Clear the dirty flag in the whole TREE */
 static void tree_clean(struct tree *tree) {
+    if (tree->dirty) {
+        list_for_each(c, tree->children)
+            tree_clean(c);
+    }
     tree->dirty = 0;
-    list_for_each(c, tree->children)
-        tree_clean(c);
+}
+
+static struct tree *tree_child(struct tree *tree, const char *label) {
+    if (tree == NULL)
+        return NULL;
+
+    list_for_each(child, tree->children) {
+        if (streqv(label, child->label))
+            return child;
+    }
+    return NULL;
+}
+
+/* Get first existing child or create one */
+static struct tree *tree_child_cr(struct tree *tree, const char *label) {
+    static struct tree *child = NULL;
+
+    if (tree == NULL)
+        return NULL;
+
+    child = tree_child(tree, label);
+    if (child == NULL) {
+        char *l = strdup(label);
+        if (l == NULL)
+            return NULL;
+        child = tree_append(tree, l, NULL);
+    }
+    return child;
+}
+
+static struct tree *tree_path_cr(struct tree *tree, int n, ...) {
+    va_list ap;
+
+    va_start(ap, n);
+    for (int i=0; i < n; i++) {
+        const char *l = va_arg(ap, const char *);
+        tree = tree_child_cr(tree, l);
+    }
+    va_end(ap);
+    return tree;
+}
+
+static int tree_set_value(struct tree *tree, const char *value) {
+    if (tree->value != NULL) {
+        free(tree->value);
+        tree->value = NULL;
+    }
+    if (value != NULL) {
+        tree->value = strdup(value);
+        if (tree->value == NULL)
+            return -1;
+    }
+    tree_mark_dirty(tree);
+    return 0;
+}
+
+/* Parse a path expression. Report errors in /augeas/pathx/error */
+static struct pathx *parse_user_pathx(const struct augeas *aug,
+                                      bool need_nodeset,
+                                      const char *path) {
+    struct pathx *result;
+    int    pos;
+
+    if (pathx_parse(aug->origin, path, need_nodeset, aug->symtab, &result)
+        == PATHX_NOERROR)
+        return result;
+
+    struct tree *error =
+        tree_path_cr(aug->origin, 3, s_augeas, s_pathx, s_error);
+    if (error == NULL)
+        return NULL;
+    tree_set_value(error, pathx_error(result, NULL, &pos));
+
+    struct tree *tpos = tree_child_cr(error, s_pos);
+    if (tpos == NULL)
+        return NULL;
+
+    free(tpos->value);
+    if (ALLOC_N(tpos->value, strlen(path) + 4) >= 0) {
+        strncpy(tpos->value, path, pos);
+        strcat(tpos->value, "|=|");
+        strcat(tpos->value, path + pos);
+    }
+    return NULL;
 }
 
 static const char *init_root(const char *root0) {
@@ -118,6 +159,8 @@ static const char *init_root(const char *root0) {
     if (root0 == NULL)
         root0 = "/";
     root = strdup(root0);
+    if (root == NULL)
+        return NULL;
     if (root[strlen(root)-1] != SEP) {
         if (REALLOC_N(root, strlen(root) + 2) == -1) {
             FREE(root);
@@ -128,6 +171,54 @@ static const char *init_root(const char *root0) {
     return root;
 }
 
+struct tree *tree_append(struct tree *parent,
+                         char *label, char *value) {
+    struct tree *result = make_tree(label, value, parent, NULL);
+    if (result != NULL)
+        list_append(parent->children, result);
+    return result;
+}
+
+static struct tree *tree_from_transform(struct augeas *aug,
+                                        const char *modname,
+                                        struct transform *xfm) {
+    struct tree *meta = tree_child_cr(aug->origin, s_augeas);
+    struct tree *load = NULL, *txfm = NULL;
+    char *m = NULL, *q = NULL;
+
+    if (meta == NULL)
+        goto error;
+
+    load = tree_child_cr(meta, s_load);
+    if (load == NULL)
+            goto error;
+    if (modname != NULL) {
+        m = strdup(modname);
+        if (m == NULL)
+            goto error;
+    } else {
+        m = strdup("_");
+    }
+    txfm = tree_append(load, m, NULL);
+    if (txfm == NULL)
+        goto error;
+
+    if (asprintf(&q, "@%s", modname) < 0)
+        goto error;
+
+    m = strdup("lens");
+    tree_append(txfm, m, q);
+    list_for_each(f, xfm->filter) {
+        char *glob = strdup(f->glob->str);
+        char *l = f->include ? strdup("incl") : strdup("excl");
+        tree_append(txfm, l, glob);
+    }
+    return txfm;
+ error:
+    tree_unlink(txfm);
+    return NULL;
+}
+
 struct augeas *aug_init(const char *root, const char *loadpath,
                         unsigned int flags) {
     struct augeas *result;
@@ -136,7 +227,8 @@ struct augeas *aug_init(const char *root, const char *loadpath,
     if (tree_root == NULL)
         return NULL;
 
-    CALLOC(result, 1);
+    if (ALLOC(result) < 0)
+        goto error;
     result->origin = make_tree_origin(tree_root);
     if (result->origin == NULL) {
         free_tree(tree_root);
@@ -147,7 +239,7 @@ struct augeas *aug_init(const char *root, const char *loadpath,
 
     result->root = init_root(root);
 
-    result->origin->children->label = strdup(P_ROOT);
+    result->origin->children->label = strdup(s_augeas);
 
     result->modpathz = NULL;
     result->nmodpath = 0;
@@ -161,6 +253,7 @@ struct augeas *aug_init(const char *root, const char *loadpath,
     }
     if (!(flags & AUG_NO_STDINC)) {
         argz_add(&result->modpathz, &result->nmodpath, AUGEAS_LENS_DIR);
+        argz_add(&result->modpathz, &result->nmodpath, AUGEAS_LENS_DIST_DIR);
     }
     /* Clean up trailing slashes */
     if (result->nmodpath > 0) {
@@ -191,13 +284,20 @@ struct augeas *aug_init(const char *root, const char *loadpath,
        AUGEAS_META_ROOT getting changed. */
     aug_set(result, AUGEAS_META_ROOT, result->root);
 
+    for (int i=0; i < ARRAY_CARDINALITY(static_nodes); i++)
+        aug_set(result, static_nodes[i][0], static_nodes[i][1]);
+
     if (flags & AUG_SAVE_NEWFILE) {
-        aug_set(result, AUGEAS_META_SAVE_MODE, "newfile");
+        aug_set(result, AUGEAS_META_SAVE_MODE, AUG_SAVE_NEWFILE_TEXT);
     } else if (flags & AUG_SAVE_BACKUP) {
-        aug_set(result, AUGEAS_META_SAVE_MODE, "backup");
+        aug_set(result, AUGEAS_META_SAVE_MODE, AUG_SAVE_BACKUP_TEXT);
+    } else if (flags & AUG_SAVE_NOOP) {
+        aug_set(result, AUGEAS_META_SAVE_MODE, AUG_SAVE_NOOP_TEXT);
     } else {
-        aug_set(result, AUGEAS_META_SAVE_MODE, "overwrite");
+        aug_set(result, AUGEAS_META_SAVE_MODE, AUG_SAVE_OVERWRITE_TEXT);
     }
+    /* Make sure we always have /files */
+    aug_set(result, AUGEAS_FILES_TREE, NULL);
 
     if (interpreter_init(result) == -1)
         goto error;
@@ -206,11 +306,12 @@ struct augeas *aug_init(const char *root, const char *loadpath,
         struct transform *xform = modl->autoload;
         if (xform == NULL)
             continue;
-        transform_load(result, xform);
-    }
-    list_for_each(tree, result->origin->children) {
-        tree_clean(tree);
+        tree_from_transform(result, modl->name, xform);
     }
+    if (!(result->flags & AUG_NO_LOAD))
+        if (aug_load(result) < 0)
+            goto error;
+
     return result;
 
  error:
@@ -218,73 +319,139 @@ struct augeas *aug_init(const char *root, const char *loadpath,
     return NULL;
 }
 
+static void tree_unlink_children(struct tree *tree) {
+    if (tree == NULL)
+        return;
+
+    while (tree->children != NULL)
+        tree_unlink(tree->children);
+}
+
+int aug_load(struct augeas *aug) {
+    struct tree *meta = tree_child_cr(aug->origin, s_augeas);
+    struct tree *meta_files = tree_child_cr(meta, s_files);
+    struct tree *files = tree_child_cr(aug->origin, s_files);
+    struct tree *load = tree_child_cr(meta, s_load);
+
+    if (load == NULL)
+        return -1;
+
+    tree_unlink_children(meta_files);
+    tree_unlink_children(files);
+
+    list_for_each(xfm, load->children) {
+        if (transform_validate(aug, xfm) == 0)
+            transform_load(aug, xfm);
+    }
+    tree_clean(aug->origin);
+    return 0;
+}
+
 int aug_get(const struct augeas *aug, const char *path, const char **value) {
-    struct path *p = make_path(aug->origin->children, path);
+    struct pathx *p;
     struct tree *match;
     int r;
 
+    p = parse_user_pathx((struct augeas *) aug, true, path);
     if (p == NULL)
         return -1;
 
     if (value != NULL)
         *value = NULL;
 
-    r = path_find_one(p, &match);
+    r = pathx_find_one(p, &match);
     if (r == 1 && value != NULL)
         *value = match->value;
-    free_path(p);
+    free_pathx(p);
 
     return r;
 }
 
-struct tree *tree_set(struct tree *root, const char *path, const char *value) {
+int aug_defvar(augeas *aug, const char *name, const char *expr) {
+    struct pathx *p = NULL;
+    int result = -1;
+
+    if (expr == NULL) {
+        result = pathx_symtab_undefine(&(aug->symtab), name);
+    } else {
+        p = parse_user_pathx((struct augeas *) aug, false, expr);
+        if (p == NULL)
+            goto done;
+        result = pathx_symtab_define(&(aug->symtab), name, p);
+    }
+ done:
+    free_pathx(p);
+    return result;
+}
+
+int aug_defnode(augeas *aug, const char *name, const char *expr,
+                const char *value, int *created) {
+    struct pathx *p;
+    int result = -1;
+    int r, cr;
+    struct tree *tree;
+
+    if (expr == NULL)
+        return -1;
+    if (created == NULL)
+        created = &cr;
+
+    p = parse_user_pathx((struct augeas *) aug, false, expr);
+    if (p == NULL)
+        goto done;
+
+    r = pathx_expand_tree(p, &tree);
+    if (r < 0)
+        goto done;
+
+    *created = r > 0;
+    if (*created) {
+        r = tree_set_value(tree, value);
+        if (r < 0)
+            goto done;
+    }
+
+    result = pathx_symtab_define(&(aug->symtab), name, p);
+
+ done:
+    free_pathx(p);
+    return result;
+}
+
+struct tree *tree_set(struct pathx *p, const char *value) {
     struct tree *tree;
-    struct path *p = make_path(root, path);
     int r;
 
-    if (p == NULL)
-        goto error;
-
-    r = path_expand_tree(p, &tree);
+    r = pathx_expand_tree(p, &tree);
     if (r == -1)
-        goto error;
-    free_path(p);
+        return NULL;
 
-    if (tree->value != NULL) {
-        free(tree->value);
-        tree->value = NULL;
-    }
-    if (value != NULL) {
-        tree->value = strdup(value);
-        if (tree->value == NULL)
-            goto error;
-    }
-    tree->dirty = 1;
+    r = tree_set_value(tree, value);
+    if (r < 0)
+        return NULL;
     return tree;
- error:
-    free_path(p);
-    return NULL;
 }
 
 int aug_set(struct augeas *aug, const char *path, const char *value) {
-    return tree_set(aug->origin->children, path, value) == NULL ? -1 : 0;
+    struct pathx *p;
+    int result;
+
+    p = parse_user_pathx(aug, true, path);
+    if (p == NULL)
+        return -1;
+
+    result = tree_set(p, value) == NULL ? -1 : 0;
+    free_pathx(p);
+    return result;
 }
 
-int tree_insert(struct tree *origin, const char *path, const char *label,
-                int before) {
-    assert(origin->parent == origin);
-
-    struct path *p = NULL;
+int tree_insert(struct pathx *p, const char *label, int before) {
     struct tree *new = NULL, *match;
 
     if (strchr(label, SEP) != NULL)
         return -1;
 
-    p = make_path(origin->children, path);
-    if (p == NULL)
-        goto error;
-
-    if (path_find_one(p, &match) != 1)
+    if (pathx_find_one(p, &match) != 1)
         goto error;
 
     new = make_tree(strdup(label), NULL, match->parent, NULL);
@@ -297,24 +464,31 @@ int tree_insert(struct tree *origin, const char *path, const char *label,
         new->next = match->next;
         match->next = new;
     }
-    free_path(p);
     return 0;
  error:
     free_tree(new);
-    free_path(p);
     return -1;
 }
 
 int aug_insert(struct augeas *aug, const char *path, const char *label,
                int before) {
-    return tree_insert(aug->origin, path, label, before);
+    struct pathx *p = NULL;
+    int result = -1;
+
+    p = parse_user_pathx(aug, true, path);
+    if (p == NULL)
+        goto done;
+
+    result = tree_insert(p, label, before);
+ done:
+    free_pathx(p);
+    return result;
 }
 
 struct tree *make_tree(char *label, char *value, struct tree *parent,
                        struct tree *children) {
     struct tree *tree;
-    CALLOC(tree, 1);
-    if (tree == NULL)
+    if (ALLOC(tree) < 0)
         return NULL;
 
     tree->label = label;
@@ -323,7 +497,10 @@ struct tree *make_tree(char *label, char *value, struct tree *parent,
     tree->children = children;
     list_for_each(c, tree->children)
         c->parent = tree;
-    tree->dirty = 1;
+    if (parent != NULL)
+        tree_mark_dirty(tree);
+    else
+        tree->dirty = 1;
     return tree;
 }
 
@@ -363,67 +540,76 @@ int free_tree(struct tree *tree) {
     return cnt;
 }
 
-int tree_rm(struct tree *origin, const char *path) {
-    assert(origin->parent == origin);
-    assert(origin->next == NULL);
+int tree_unlink(struct tree *tree) {
+    int result = 0;
 
-    struct path *p = NULL;
-    struct tree *root = origin->children;
+    assert (tree->parent != NULL);
+    list_remove(tree, tree->parent->children);
+    tree_mark_dirty(tree->parent);
+    result = free_tree(tree->children) + 1;
+    free_tree_node(tree);
+    return result;
+}
+
+int tree_rm(struct pathx *p) {
     struct tree *tree, **del;
     int cnt = 0, ndel = 0, i;
 
-    p = make_path(root, path);
-    if (p == NULL)
-        return -1;
-
-    for (tree = path_first(p); tree != NULL; tree = path_next(p, tree)) {
+    for (tree = pathx_first(p); tree != NULL; tree = pathx_next(p)) {
         if (! TREE_HIDDEN(tree))
             ndel += 1;
     }
 
-    if (ndel == 0) {
-        free_path(p);
+    if (ndel == 0)
         return 0;
-    }
 
     if (ALLOC_N(del, ndel) < 0) {
         free(del);
         return -1;
     }
 
-    for (i = 0, tree = path_first(p); tree != NULL; tree = path_next(p, tree)) {
+    for (i = 0, tree = pathx_first(p); tree != NULL; tree = pathx_next(p)) {
         if (TREE_HIDDEN(tree))
             continue;
+        pathx_symtab_remove_descendants(p, tree);
         del[i] = tree;
         i += 1;
     }
-    free_path(p);
 
-    for (i = 0; i < ndel; i++) {
-        assert (del[i]->parent != NULL);
-        list_remove(del[i], del[i]->parent->children);
-        del[i]->parent->dirty = 1;
-        cnt += free_tree(del[i]->children) + 1;
-        free_tree_node(del[i]);
-    }
+    for (i = 0; i < ndel; i++)
+        cnt += tree_unlink(del[i]);
     free(del);
 
     return cnt;
 }
 
 int aug_rm(struct augeas *aug, const char *path) {
-    return tree_rm(aug->origin, path);
+    struct pathx *p = NULL;
+    int result;
+
+    p = parse_user_pathx(aug, true, path);
+    if (p == NULL)
+        return -1;
+
+    result = tree_rm(p);
+    free_pathx(p);
+
+    return result;
 }
 
-int aug_tree_replace(struct augeas *aug, const char *path, struct tree *sub) {
+int tree_replace(struct tree *origin, const char *path, struct tree *sub) {
     struct tree *parent;
+    struct pathx *p = NULL;
     int r;
 
-    r = aug_rm(aug, path);
+    if (pathx_parse(origin, path, true, NULL, &p) != PATHX_NOERROR)
+        goto error;
+
+    r = tree_rm(p);
     if (r == -1)
         goto error;
 
-    parent = tree_set(aug->origin->children, path, NULL);
+    parent = tree_set(p, NULL);
     if (parent == NULL)
         goto error;
 
@@ -431,27 +617,34 @@ int aug_tree_replace(struct augeas *aug, const char *path, struct tree *sub) {
     list_for_each(s, sub) {
         s->parent = parent;
     }
+    if (sub->dirty)
+        tree_mark_dirty(parent);
+    free_pathx(p);
     return 0;
  error:
+    free_pathx(p);
     return -1;
 }
 
 int aug_mv(struct augeas *aug, const char *src, const char *dst) {
-    struct tree *root = aug->origin->children;
-    struct path *s = make_path(root, src);
-    struct path *d = make_path(root, dst);
+    struct pathx *s = NULL, *d = NULL;
     struct tree *ts, *td, *t;
     int r, ret;
 
     ret = -1;
-    if (s == NULL || d == NULL)
+    s = parse_user_pathx(aug, true, src);
+    if (s == NULL)
         goto done;
 
-    r = path_find_one(s, &ts);
+    d = parse_user_pathx(aug, true, dst);
+    if (d == NULL)
+        goto done;
+
+    r = pathx_find_one(s, &ts);
     if (r != 1)
         goto done;
 
-    r = path_expand_tree(d, &td);
+    r = pathx_expand_tree(d, &td);
     if (r == -1)
         goto done;
 
@@ -475,23 +668,18 @@ int aug_mv(struct augeas *aug, const char *src, const char *dst) {
     ts->value = NULL;
     ts->children = NULL;
 
-
-    list_remove(ts, ts->parent->children);
-
-    ts->parent->dirty = 1;
-    td->dirty = 1;
-
-    free_tree(ts);
+    tree_unlink(ts);
+    tree_mark_dirty(td);
 
     ret = 0;
  done:
-    free_path(s);
-    free_path(d);
+    free_pathx(s);
+    free_pathx(d);
     return ret;
 }
 
 int aug_match(const struct augeas *aug, const char *pathin, char ***matches) {
-    struct path *p = NULL;
+    struct pathx *p = NULL;
     struct tree *tree;
     int cnt = 0;
 
@@ -502,36 +690,33 @@ int aug_match(const struct augeas *aug, const char *pathin, char ***matches) {
         pathin = "/*";
     }
 
-    p = make_path(aug->origin->children, pathin);
+    p = parse_user_pathx((struct augeas *) aug, true, pathin);
     if (p == NULL)
         return -1;
 
-    for (tree = path_first(p); tree != NULL; tree = path_next(p, tree)) {
+    for (tree = pathx_first(p); tree != NULL; tree = pathx_next(p)) {
         if (! TREE_HIDDEN(tree))
             cnt += 1;
     }
-    free_path(p);
-    p = NULL;
 
     if (matches == NULL)
-        return cnt;
+        goto done;
 
-    CALLOC(*matches, cnt);
-    if (*matches == NULL)
+    if (ALLOC_N(*matches, cnt) < 0)
         goto error;
 
-    p = make_path(aug->origin->children, pathin);
     int i = 0;
-    for (tree = path_first(p); tree != NULL; tree = path_next(p, tree)) {
+    for (tree = pathx_first(p); tree != NULL; tree = pathx_next(p)) {
         if (TREE_HIDDEN(tree))
             continue;
-        (*matches)[i] = format_path(tree);
+        (*matches)[i] = path_of_tree(tree);
         if ((*matches)[i] == NULL) {
             goto error;
         }
         i += 1;
     }
-    free_path(p);
+ done:
+    free_pathx(p);
     return cnt;
 
  error:
@@ -542,31 +727,33 @@ int aug_match(const struct augeas *aug, const char *pathin, char ***matches) {
             free(*matches);
         }
     }
-    free_path(p);
+    free_pathx(p);
     return -1;
 }
 
-static int tree_save(struct augeas *aug, struct tree *tree, const char *path,
-                     int *count) {
+static int tree_save(struct augeas *aug, struct tree *tree,
+                     const char *path) {
     int result = 0;
+    struct tree *meta = tree_child_cr(aug->origin, s_augeas);
+    struct tree *load = tree_child_cr(meta, s_load);
+
     // FIXME: We need to detect subtrees that aren't saved by anything
-    aug_rm(aug, AUGEAS_EVENTS_SAVED);
+
+    if (load == NULL)
+        return -1;
 
     list_for_each(t, tree) {
         if (t->dirty) {
             char *tpath = NULL;
-            struct transform *transform = NULL;
+            struct tree *transform = NULL;
             if (asprintf(&tpath, "%s/%s", path, t->label) == -1) {
                 result = -1;
                 continue;
             }
-            list_for_each(modl, aug->modules) {
-                struct transform *xform = modl->autoload;
-                if (xform == NULL)
-                    continue;
-                if (transform_applies(xform, tpath)) {
-                    if (transform == NULL || transform == xform) {
-                        transform = xform;
+            list_for_each(xfm, load->children) {
+                if (transform_applies(xfm, tpath)) {
+                    if (transform == NULL || transform == xfm) {
+                        transform = xfm;
                     } else {
                         FIXME("Multiple transforms for %s", path);
                         result = -1;
@@ -577,23 +764,8 @@ static int tree_save(struct augeas *aug, struct tree *tree, const char *path,
                 int r = transform_save(aug, transform, tpath, t);
                 if (r == -1)
                     result = -1;
-                else if (r > 0) {
-                    char *sp;
-                    *count += 1;
-                    if (asprintf(&sp, AUGEAS_EVENTS_SAVED "[%d]",
-                                 *count) < 0) {
-                        free(tpath);
-                        return -1;
-                    }
-                    if (aug_set(aug, sp, tpath) < 0) {
-                        free(tpath);
-                        free(sp);
-                        return -1;
-                    }
-                    free(sp);
-                }
             } else {
-                if (tree_save(aug, t->children, tpath, count) == -1)
+                if (tree_save(aug, t->children, tpath) == -1)
                     result = -1;
             }
             free(tpath);
@@ -602,28 +774,102 @@ static int tree_save(struct augeas *aug, struct tree *tree, const char *path,
     return result;
 }
 
-int aug_save(struct augeas *aug) {
-    int ret = 0;
-    struct tree *files;
-    struct path *p = make_path(aug->origin->children, AUGEAS_FILES_TREE);
+/* Reset the flags based on what is set in the tree. */
+static int update_save_flags(struct augeas *aug) {
+    const char *savemode ;
+    int noop = 0 ;
 
-    if (p == NULL || path_find_one(p, &files) != 1) {
-        free_path(p);
+    aug_get(aug, AUGEAS_META_SAVE_MODE, &savemode);
+    if (savemode == NULL)
+        return -1;
+
+    aug->flags &= ~(AUG_SAVE_BACKUP|AUG_SAVE_NEWFILE|AUG_SAVE_NOOP);
+    if (STREQ(savemode, AUG_SAVE_NEWFILE_TEXT)) {
+        aug->flags |= AUG_SAVE_NEWFILE;
+    } else if (STREQ(savemode, AUG_SAVE_BACKUP_TEXT)) {
+        aug->flags |= AUG_SAVE_BACKUP;
+    } else if (STREQ(savemode, AUG_SAVE_NOOP_TEXT)) {
+        aug->flags |= AUG_SAVE_NOOP ;
+        noop = 1 ;
+    } else if (STRNEQ(savemode, AUG_SAVE_OVERWRITE_TEXT)) {
         return -1;
     }
-    free_path(p);
 
-    list_for_each(t, aug->origin->children) {
-        tree_propagate_dirty(t);
+    return 0;
+}
+
+static int unlink_removed_files(struct augeas *aug,
+                                struct tree *files, struct tree *meta) {
+    /* Find all nodes that correspond to a file and might have to be
+     * unlinked. A node corresponds to a file if it has a child labelled
+     * 'path', and we only consider it if there are no errors associated
+     * with it */
+    static const char *const file_nodes =
+        "descendant-or-self::*[path][count(error) = 0]";
+
+    int result = 0;
+
+    if (! files->dirty)
+        return 0;
+
+    for (struct tree *tm = meta->children; tm != NULL;) {
+        struct tree *tf = tree_child(files, tm->label);
+        struct tree *next = tm->next;
+        if (tf == NULL) {
+            /* Unlink all files in tm */
+            struct pathx *px = NULL;
+            if (pathx_parse(tm, file_nodes, true, NULL, &px)
+                != PATHX_NOERROR) {
+                result = -1;
+                continue;
+            }
+            for (struct tree *t = pathx_first(px);
+                 t != NULL;
+                 t = pathx_next(px)) {
+                remove_file(aug, t);
+            }
+            free_pathx(px);
+        } else if (tf->dirty && ! tree_child(tm, "path")) {
+            if (unlink_removed_files(aug, tf, tm) < 0)
+                result = -1;
+        }
+        tm = next;
     }
+    return result;
+}
+
+int aug_save(struct augeas *aug) {
+    int ret = 0;
+    struct tree *meta = tree_child_cr(aug->origin, s_augeas);
+    struct tree *meta_files = tree_child_cr(meta, s_files);
+    struct tree *files = tree_child_cr(aug->origin, s_files);
+    struct tree *load = tree_child_cr(meta, s_load);
+
+    if (update_save_flags(aug) < 0)
+        return -1;
+
+    if (files == NULL || meta == NULL || load == NULL)
+        return -1;
+
+    aug_rm(aug, AUGEAS_EVENTS_SAVED);
+
+    list_for_each(xfm, load->children)
+        transform_validate(aug, xfm);
+
     if (files->dirty) {
-        int count = 0;
         list_for_each(t, files->children) {
-            if (tree_save(aug, t, AUGEAS_FILES_TREE, &count) == -1)
+            if (tree_save(aug, t, AUGEAS_FILES_TREE) == -1)
+                ret = -1;
+        }
+        /* Remove files whose entire subtree was removed. */
+        if (meta_files != NULL) {
+            if (unlink_removed_files(aug, files, meta_files) < 0)
                 ret = -1;
         }
     }
-    tree_clean(aug->origin->children);
+    if (!(aug->flags & AUG_SAVE_NOOP)) {
+        tree_clean(aug->origin);
+    }
     return ret;
 }
 
@@ -675,22 +921,16 @@ static int print_rec(FILE *out, struct tree *start, const char *ppath,
     return -1;
 }
 
-int print_tree(const struct tree *start, FILE *out, const char *pathin,
-               int pr_hidden) {
-
-    struct path *p = make_path(start, pathin);
+static int print_tree(FILE *out, struct pathx *p, int pr_hidden) {
     char *path = NULL;
     struct tree *tree;
     int r;
 
-    if (p == NULL)
-        return -1;
-
-    for (tree = path_first(p); tree != NULL; tree = path_next(p, tree)) {
+    for (tree = pathx_first(p); tree != NULL; tree = pathx_next(p)) {
         if (TREE_HIDDEN(tree) && ! pr_hidden)
             continue;
 
-        path = format_path(tree);
+        path = path_of_tree(tree);
         if (path == NULL)
             goto error;
         r = print_one(out, path, tree->value);
@@ -702,18 +942,40 @@ int print_tree(const struct tree *start, FILE *out, const char *pathin,
         free(path);
         path = NULL;
     }
-    free_path(p);
     return 0;
  error:
     free(path);
     return -1;
 }
 
+int dump_tree(FILE *out, struct tree *tree) {
+    struct pathx *p;
+    int result;
+
+    if (pathx_parse(tree, "/*", true, NULL, &p) != PATHX_NOERROR)
+        return -1;
+
+    result = print_tree(out, p, 1);
+    free_pathx(p);
+    return result;
+}
+
 int aug_print(const struct augeas *aug, FILE *out, const char *pathin) {
+    struct pathx *p;
+    int result;
+
     if (pathin == NULL || strlen(pathin) == 0) {
         pathin = "/*";
     }
-    return print_tree(aug->origin->children, out, pathin, 0);
+
+    p = parse_user_pathx((struct augeas *) aug, true, pathin);
+    if (p == NULL)
+        return -1;
+
+    result = print_tree(out, p, 0);
+    free_pathx(p);
+
+    return result;
 }
 
 void aug_close(struct augeas *aug) {
@@ -723,6 +985,7 @@ void aug_close(struct augeas *aug) {
     unref(aug->modules, module);
     free((void *) aug->root);
     free(aug->modpathz);
+    free_symtab(aug->symtab);
     free(aug);
 }
 
@@ -740,6 +1003,7 @@ int tree_equal(const struct tree *t1, const struct tree *t2) {
     return t1 == t2;
 }
 
+
 /*
  * Local variables:
  *  indent-tabs-mode: nil
