@@ -47,15 +47,20 @@ static unsigned int flags = AUG_NONE;
 const char *root = NULL;
 char *loadpath = NULL;
 
-static char *cleanpath(char *path) {
+
+static char *cleanstr(char *path, const char sep) {
     if (path == NULL || strlen(path) == 0)
         return path;
     char *e = path + strlen(path) - 1;
-    while (e != path && (*e == SEP || isspace(*e)))
+    while (e >= path && (*e == sep || isspace(*e)))
         *e-- = '\0';
     return path;
 }
 
+static char *cleanpath(char *path) {
+    return cleanstr(path, SEP);
+}
+
 /*
  * Dup PATH and split it into a directory and basename. The returned value
  * points to the copy of PATH. Adding strlen(PATH)+1 to it gives the
@@ -559,6 +564,7 @@ static void usage(void) {
     fprintf(stderr, "  -I, --include DIR  search DIR for modules; can be given mutiple times\n");
     fprintf(stderr, "  --nostdinc         do not search the builtin default directories for modules\n");
     fprintf(stderr, "  --noload           do not load any files into the tree on startup\n");
+    fprintf(stderr, "  --noautoload       do not autoload modules from the search path\n");
 
     exit(EXIT_FAILURE);
 }
@@ -568,7 +574,8 @@ static void parse_opts(int argc, char **argv) {
     size_t loadpathlen = 0;
     enum {
         VAL_NO_STDINC = CHAR_MAX + 1,
-        VAL_NO_LOAD = VAL_NO_STDINC + 1
+        VAL_NO_LOAD = VAL_NO_STDINC + 1,
+        VAL_NO_AUTOLOAD = VAL_NO_LOAD + 1
     };
     struct option options[] = {
         { "help",      0, 0, 'h' },
@@ -579,6 +586,7 @@ static void parse_opts(int argc, char **argv) {
         { "include",   1, 0, 'I' },
         { "nostdinc",  0, 0, VAL_NO_STDINC },
         { "noload",    0, 0, VAL_NO_LOAD },
+        { "noautoload", 0, 0, VAL_NO_AUTOLOAD },
         { 0, 0, 0, 0}
     };
     int idx;
@@ -609,6 +617,9 @@ static void parse_opts(int argc, char **argv) {
         case VAL_NO_LOAD:
             flags |= AUG_NO_LOAD;
             break;
+        case VAL_NO_AUTOLOAD:
+            flags |= AUG_NO_MODL_AUTOLOAD;
+            break;
         default:
             usage();
             break;
@@ -619,18 +630,26 @@ static void parse_opts(int argc, char **argv) {
 
 static int main_loop(void) {
     static const int maxargs = 3;
-    char *line;
+    char *line = NULL;
     char *cmd, *args[maxargs];
     int ret = 0;
+    size_t len = 0;
 
     while(1) {
         char *dup_line;
 
-        line = readline("augtool> ");
+        if (isatty(fileno(stdin))) {
+            line = readline("augtool> ");
+        } else if (getline(&line, &len, stdin) == -1) {
+            return ret;
+        }
+        cleanstr(line, '\n');
         if (line == NULL) {
             printf("\n");
             return ret;
         }
+        if (line[0] == '#')
+            continue;
 
         dup_line = strdup(line);
         if (dup_line == NULL) {
@@ -644,7 +663,8 @@ static int main_loop(void) {
             r = run_command(cmd, maxargs, args);
             if (r < 0)
                 ret = -1;
-            add_history(line);
+            if (isatty(fileno(stdin)))
+                add_history(line);
         }
         free(dup_line);
     }
