@@ -23,9 +23,12 @@
 #include <config.h>
 #include <stdio.h>
 #include <stdarg.h>
+#include <errno.h>
+#include <stdlib.h>
 
 #include "syntax.h"
 #include "memory.h"
+#include "transform.h"
 
 #define UNIMPL_BODY(name)                       \
     {                                           \
@@ -106,12 +109,32 @@ static void exn_print_tree(struct value *exn, struct tree *tree) {
     struct memstream ms;
 
     init_memstream(&ms);
-    print_tree(tree, ms.stream, "/*", 1);
+    dump_tree(ms.stream, tree);
     close_memstream(&ms);
     exn_printf_line(exn, "%s", ms.buf);
     FREE(ms.buf);
 }
 
+static struct value *make_pathx_exn(struct info *info, struct pathx *p) {
+    struct value *v;
+    char *msg;
+    const char *txt;
+    int pos;
+
+    msg = strdup(pathx_error(p, &txt, &pos));
+    if (msg == NULL)
+        return NULL;
+
+    v = make_exn_value(ref(info), "syntax error in path expression: %s", msg);
+    if (ALLOC_N(msg, strlen(txt) + 4) >= 0) {
+        strncpy(msg, txt, pos);
+        strcat(msg, "|=|");
+        strcat(msg, txt + pos);
+        exn_add_lines(v, 1, msg);
+    }
+    return v;
+}
+
 /* V_LENS -> V_STRING -> V_TREE */
 static struct value *lens_get(struct info *info, struct value *l,
                               struct value *str) {
@@ -126,12 +149,15 @@ static struct value *lens_get(struct info *info, struct value *l,
         v = make_value(V_TREE, ref(info));
         v->origin = make_tree_origin(tree);
     } else {
-        tree = make_tree_origin(tree);
+        struct tree *t = make_tree_origin(tree);
+        if (t == NULL)
+            free_tree(tree);
+        tree = t;
 
         v = make_exn_lns_error(info, err, text);
         if (tree != NULL) {
             exn_printf_line(v, "Tree generated so far:");
-            exn_print_tree(v, tree->children);
+            exn_print_tree(v, tree);
             free_tree(tree);
         }
         free_lns_error(err);
@@ -178,24 +204,35 @@ static struct value *tree_set_glue(struct info *info, struct value *path,
     assert(tree->tag == V_TREE);
 
     struct tree *fake = NULL;
+    struct pathx *p = NULL;
+    struct value *result = NULL;
 
     if (tree->origin->children == NULL) {
         tree->origin->children = make_tree(NULL, NULL, tree->origin, NULL);
         fake = tree->origin->children;
     }
 
-    if (tree_set(tree->origin->children, path->string->str,
-                 val->string->str) == NULL) {
-        return make_exn_value(ref(info),
-                              "Tree set of %s to '%s' failed",
-                              path->string->str, val->string->str);
+    if (pathx_parse(tree->origin, path->string->str, true, NULL, &p)
+        != PATHX_NOERROR) {
+        result = make_pathx_exn(ref(info), p);
+        goto done;
+    }
+
+    if (tree_set(p, val->string->str) == NULL) {
+        result = make_exn_value(ref(info),
+                                "Tree set of %s to '%s' failed",
+                                path->string->str, val->string->str);
+        goto done;
     }
     if (fake != NULL) {
         list_remove(fake, tree->origin->children);
         free_tree(fake);
     }
+    result = ref(tree);
 
-    return ref(tree);
+ done:
+    free_pathx(p);
+    return result;
 }
 
 static struct value *tree_insert_glue(struct info *info, struct value *label,
@@ -209,15 +246,27 @@ static struct value *tree_insert_glue(struct info *info, struct value *label,
     assert(tree->tag == V_TREE);
 
     int r;
-    r = tree_insert(tree->origin, path->string->str,
-                    label->string->str, before);
-    if (r != 0) {
-        return make_exn_value(ref(info),
-                              "Tree insert of %s at %s failed",
-                              label->string->str, path->string->str);
+    struct pathx *p = NULL;
+    struct value *result = NULL;
+
+    if (pathx_parse(tree->origin, path->string->str, true, NULL, &p)
+        != PATHX_NOERROR) {
+        result = make_pathx_exn(ref(info), p);
+        goto done;
     }
 
-    return ref(tree);
+    r = tree_insert(p, label->string->str, before);
+    if (r != 0) {
+        result = make_exn_value(ref(info),
+                                "Tree insert of %s at %s failed",
+                                label->string->str, path->string->str);
+        goto done;
+    }
+
+    result = ref(tree);
+ done:
+    free_pathx(p);
+    return result;
 }
 
 /* Insert after */
@@ -243,11 +292,25 @@ static struct value *tree_rm_glue(struct info *info,
     // need to copy TREE first
     assert(path->tag == V_STRING);
     assert(tree->tag == V_TREE);
-    if (tree_rm(tree->origin, path->string->str) == -1) {
-        return make_exn_value(ref(info), "Tree rm of %s failed",
-                              path->string->str);
+
+    struct pathx *p = NULL;
+    struct value *result = NULL;
+
+    if (pathx_parse(tree->origin, path->string->str, true, NULL, &p)
+        != PATHX_NOERROR) {
+        result = make_pathx_exn(ref(info), p);
+        goto done;
     }
-    return ref(tree);
+
+    if (tree_rm(p) == -1) {
+        result = make_exn_value(ref(info), "Tree rm of %s failed",
+                                path->string->str);
+        goto done;
+    }
+    result = ref(tree);
+ done:
+    free_pathx(p);
+    return result;
 }
 
 /* V_STRING -> V_STRING */
@@ -297,6 +360,32 @@ static struct value *xform_transform(struct info *info, struct value *l,
     return v;
 }
 
+static struct value *sys_getenv(struct info *info, struct value *n) {
+    assert(n->tag == V_STRING);
+    struct value *v = make_value(V_STRING, ref(info));
+    v->string = dup_string(getenv(n->string->str));
+    return v;
+}
+
+static struct value *sys_read_file(struct info *info, struct value *n) {
+    assert(n->tag == V_STRING);
+    char *str = NULL;
+
+    str = read_file(n->string->str);
+    if (str == NULL) {
+        char error_buf[1024];
+        const char *errmsg;
+        errmsg = xstrerror(errno, error_buf, sizeof(error_buf));
+        struct value *exn = make_exn_value(ref(info),
+             "reading file %s failed:", n->string->str);
+        exn_printf_line(exn, "%s", errmsg);
+        return exn;
+    }
+    struct value *v = make_value(V_STRING, ref(info));
+    v->string = make_string(str);
+    return v;
+}
+
 struct module *builtin_init(void) {
     struct module *modl = module_create("Builtin");
     define_native(modl, "gensym", 1, gensym, T_STRING, T_STRING);
@@ -324,6 +413,11 @@ struct module *builtin_init(void) {
     define_native(modl, "excl", 1, xform_excl, T_STRING, T_FILTER);
     define_native(modl, "transform", 2, xform_transform, T_LENS, T_FILTER,
                                                          T_TRANSFORM);
+    /* System functions */
+    struct module *sys = module_create("Sys");
+    modl->next = sys;
+    define_native(sys, "getenv", 1, sys_getenv, T_STRING, T_STRING);
+    define_native(sys, "read_file", 1, sys_read_file, T_STRING, T_STRING);
     return modl;
 }
 
