@@ -32,6 +32,7 @@
 #include <config.h>
 #include <limits.h>
 #include <ctype.h>
+#include <stdbool.h>
 
 #include "internal.h"
 #include "memory.h"
@@ -62,6 +63,7 @@ struct fa {
    states so that we can free the list when we need to free the state */
 struct state {
     struct state *next;
+    hash_val_t    hash;
     unsigned int  accept : 1;
     unsigned int  live : 1;
     unsigned int  reachable : 1;
@@ -113,6 +115,17 @@ static inline int bitset_get(const bitset * const bs, unsigned int bit) {
     return (bs[bit/UINT_BIT] >> bit % UINT_BIT) & 1;
 }
 
+ATTRIBUTE_PURE
+static inline bool bitset_disjoint(const bitset *const bs1,
+                                   const bitset *const bs2,
+                                   size_t nbits) {
+    for (int i=0; i < (nbits + UINT_BIT) / UINT_BIT; i++) {
+        if (bs1[i] & bs2[i])
+            return false;
+    }
+    return true;
+}
+
 static void bitset_free(bitset *bs) {
     free(bs);
 }
@@ -178,6 +191,8 @@ struct re {
 /* A map from a set of states to a state. */
 typedef hash_t state_set_hash;
 
+static hash_val_t ptr_hash(const void *p);
+
 static const int array_initial_size = 4;
 static const int array_max_expansion   = 128;
 
@@ -325,6 +340,7 @@ static struct state *make_state(void) {
     struct state *s;
     if (ALLOC(s) == -1)
         return NULL;
+    s->hash = ptr_hash(s);
     return s;
 }
 
@@ -659,18 +675,12 @@ static hash_val_t ptr_hash(const void *p) {
 typedef hash_t state_triple_hash;
 
 static hash_val_t pair_hash(const void *key) {
-    struct state *const *pair = key;
-    return ptr_hash(pair[0]) + ptr_hash(pair[1]);
+    register struct state *const *pair = key;
+    return pair[0]->hash + pair[1]->hash;
 }
 
 static int pair_cmp(const void *key1, const void *key2) {
-    struct state *const *pair1 = key1;
-    struct state *const *pair2 = key2;
-
-    if (pair1[0] < pair2[0] ||
-        (pair1[0] == pair2[0] && pair1[1] < pair2[1]))
-        return -1;
-    return pair1[1] == pair2[1] ? 0 : 1;
+    return memcmp(key1, key2, 2*sizeof(struct state *));
 }
 
 static void state_triple_node_free(hnode_t *node, ATTRIBUTE_UNUSED void *ctx) {
@@ -925,7 +935,7 @@ static hash_val_t set_hash(const void *key) {
     const struct state_set *set = key;
 
     for (int i = 0; i < set->used; i++) {
-        hash += ptr_hash(set->states[i]);
+        hash += set->states[i]->hash;
     }
     return hash;
 }
@@ -1957,6 +1967,7 @@ struct fa *fa_intersect(struct fa *fa1, struct fa *fa2) {
             }
         }
     }
+    fa->deterministic = fa1->deterministic && fa2->deterministic;
  done:
     state_set_free(worklist);
     state_triple_free(newstates);
@@ -1969,7 +1980,7 @@ struct fa *fa_intersect(struct fa *fa1, struct fa *fa2) {
     goto done;
 }
 
-int fa_contains(fa_t fa1, fa_t fa2) {
+int fa_contains(struct fa *fa1, struct fa *fa2) {
     int result = 0;
     struct state_set *worklist;  /* List of pairs of states */
     struct state_set *visited;   /* List of pairs of states */
@@ -2132,7 +2143,7 @@ struct fa *fa_overlap(struct fa *fa1, struct fa *fa2) {
     return result;
 }
 
-int fa_equals(fa_t fa1, fa_t fa2) {
+int fa_equals(struct fa *fa1, struct fa *fa2) {
     return fa_contains(fa1, fa2) && fa_contains(fa2, fa1);
 }
 
@@ -2188,7 +2199,10 @@ static char pick_char(struct trans *t) {
 /* Generate an example string for FA. Traverse all transitions and record
  * at each turn the "best" word found for that state.
  */
-char *fa_example(fa_t fa) {
+int fa_example(struct fa *fa, char **example, size_t *example_len) {
+    *example = NULL;
+    *example_len = 0;
+
     /* Sort to avoid any ambiguity because of reordering of transitions */
     sort_transition_intervals(fa);
 
@@ -2231,7 +2245,11 @@ char *fa_example(fa_t fa) {
     }
     state_set_free(path);
     state_set_free(worklist);
-    return word;
+    /* FIXME: handle embedded nul's */
+    if (word != NULL)
+        *example_len = strlen(word);
+    *example = word;
+    return 0;
 }
 
 /* Expand the automaton FA by replacing every transition s(c) -> p from
@@ -2284,42 +2302,130 @@ static struct fa *expand_alphabet(struct fa *fa, int add_marker,
     return NULL;
 }
 
+static bitset *alphabet(struct fa *fa) {
+    bitset *bs = bitset_init(UCHAR_NUM);
+
+    list_for_each(s, fa->initial) {
+        for (int i=0; i < s->tused; i++) {
+            for (uint c = s->trans[i].min; c <= s->trans[i].max; c++)
+                bitset_set(bs, c);
+        }
+    }
+    return bs;
+}
+
+static bitset *last_chars(struct fa *fa) {
+    bitset *bs = bitset_init(UCHAR_NUM);
+
+    list_for_each(s, fa->initial) {
+        for (int i=0; i < s->tused; i++) {
+            if (s->trans[i].to->accept) {
+                for (uint c = s->trans[i].min; c <= s->trans[i].max; c++)
+                    bitset_set(bs, c);
+            }
+        }
+    }
+    return bs;
+}
+
+static bitset *first_chars(struct fa *fa) {
+    bitset *bs = bitset_init(UCHAR_NUM);
+    struct state *s = fa->initial;
+
+    for (int i=0; i < s->tused; i++) {
+        for (uint c = s->trans[i].min; c <= s->trans[i].max; c++)
+            bitset_set(bs, c);
+    }
+    return bs;
+}
+
+/* Return true if F1 and F2 are known to be unambiguously concatenable
+ * according to simple heuristics. Return false if they need to be checked
+ * further to decide ambiguity */
+static bool is_splittable(struct fa *fa1, struct fa *fa2) {
+    bitset *alpha1 = NULL;
+    bitset *alpha2 = NULL;
+    bitset *last1 = NULL;
+    bitset *first2 = NULL;
+    bool result = false;
+
+    alpha2 = alphabet(fa2);
+    last1 = last_chars(fa1);
+    if (bitset_disjoint(last1, alpha2, UCHAR_NUM)) {
+        result = true;
+        goto done;
+    }
+
+    alpha1 = alphabet(fa1);
+    first2 = first_chars(fa2);
+    if (bitset_disjoint(first2, alpha1, UCHAR_NUM)) {
+        result = true;
+        goto done;
+    }
+ done:
+    bitset_free(alpha1);
+    bitset_free(alpha2);
+    bitset_free(last1);
+    bitset_free(first2);
+    return result;
+}
+
 /* This algorithm is due to Anders Moeller, and can be found in class
  * AutomatonOperations in dk.brics.grammar
  */
-char *fa_ambig_example(fa_t fa1, fa_t fa2, char **pv, char **v) {
+int fa_ambig_example(struct fa *fa1, struct fa *fa2,
+                     char **upv, size_t *upv_len,
+                     char **pv, char **v) {
     static const char X = '\001';
     static const char Y = '\002';
 
+    *upv = NULL;
+    *upv_len = 0;
+    if (pv != NULL)
+        *pv = NULL;
+    if (v != NULL)
+        *v = NULL;
+
+    if (is_splittable(fa1, fa2))
+        return 0;
+
 #define Xs "\001"
 #define Ys "\002"
+#define MPs Ys Xs "(" Xs "(.|\n))+"
+#define MSs Ys Xs "(" Xs "(.|\n))*"
+#define SPs "(" Xs "(.|\n))+" Ys Xs
+#define SSs "(" Xs "(.|\n))*" Ys Xs
     /* These could become static constants */
-    fa_t mp, ms, sp, ss;
-    fa_compile( Ys Xs "(" Xs "(.|\n))+", &mp);
-    fa_compile( Ys Xs "(" Xs "(.|\n))*", &ms);
-    fa_compile( "(" Xs "(.|\n))+" Ys Xs, &sp);
-    fa_compile("(" Xs "(.|\n))*" Ys Xs, &ss);
+    struct fa *mp, *ms, *sp, *ss;
+    fa_compile( MPs, strlen(MPs), &mp);
+    fa_compile( MSs, strlen(MSs), &ms);
+    fa_compile( SPs, strlen(SPs), &sp);
+    fa_compile( SSs, strlen(SSs), &ss);
+#undef SSs
+#undef SPs
+#undef MSs
+#undef MPs
 #undef Xs
 #undef Ys
 
-    fa_t a1f = expand_alphabet(fa1, 0, X, Y);
-    fa_t a1t = expand_alphabet(fa1, 1, X, Y);
-    fa_t a2f = expand_alphabet(fa2, 0, X, Y);
-    fa_t a2t = expand_alphabet(fa2, 1, X, Y);
+    struct fa *a1f = expand_alphabet(fa1, 0, X, Y);
+    struct fa *a1t = expand_alphabet(fa1, 1, X, Y);
+    struct fa *a2f = expand_alphabet(fa2, 0, X, Y);
+    struct fa *a2t = expand_alphabet(fa2, 1, X, Y);
 
     /* Compute b1 = ((a1f . mp) & a1t) . ms */
     concat_in_place(a1f, mp);
-    fa_t b1 = fa_intersect(a1f, a1t);
+    struct fa *b1 = fa_intersect(a1f, a1t);
     concat_in_place(b1, ms);
 
     /* Compute b2 = ss . ((sp . a2f) & a2t) */
     concat_in_place(sp, a2f);
-    fa_t b2 = fa_intersect(sp, a2t);
+    struct fa *b2 = fa_intersect(sp, a2t);
     concat_in_place(ss, b2);
     b2 = ss;
 
     /* The automaton we are really interested in */
-    fa_t amb = fa_intersect(b1, b2);
+    struct fa *amb = fa_intersect(b1, b2);
 
     /* Clean up intermediate automata */
     fa_free(sp);
@@ -2329,11 +2435,13 @@ char *fa_ambig_example(fa_t fa1, fa_t fa2, char **pv, char **v) {
     fa_free(b1);
     fa_free(b2);
 
-    char *s = fa_example(amb);
+    size_t s_len = 0;
+    char *s = NULL;
+    fa_example(amb, &s, &s_len);
     fa_free(amb);
 
     if (s == NULL)
-        return NULL;
+        return -1;
 
     char *result, *t;
     CALLOC(result, (strlen(s)-1)/2 + 1);
@@ -2355,7 +2463,10 @@ char *fa_ambig_example(fa_t fa1, fa_t fa2, char **pv, char **v) {
         *t++ = s[2*i + 1];
 
     free(s);
-    return result;
+    *upv = result;
+    if (result != NULL)
+        *upv_len = strlen(result);
+    return 0;
 }
 
 /*
@@ -2418,16 +2529,19 @@ static void free_re(struct re *re) {
     free(re);
 }
 
-int fa_compile(const char *regexp, struct fa **fa) {
+int fa_compile(const char *regexp, size_t size, struct fa **fa) {
     int ret = REG_NOERROR;
-    struct re *re = parse_regexp(&regexp, &ret);
+    struct re *re = NULL;
     *fa = NULL;
+
+    /* We don't handle embedded nul's yet */
+    if (strlen(regexp) != size)
+        return REG_ESIZE;
+
+    re = parse_regexp(&regexp, &ret);
     if (re == NULL)
         return ret;
 
-    //print_re(re);
-    //printf("\n");
-
     *fa = fa_from_re(re);
     re_unref(re);
 
@@ -2544,6 +2658,10 @@ static void parse_char_class(const char **regexp, struct re *re,
             goto error;
         }
         if (peek(regexp, "]")) {
+            if (from > to) {
+                *error = REG_ERANGE;
+                goto error;
+            }
             add_re_char(re, from, to);
             add_re_char(re, '-', '-');
             return;
@@ -2551,6 +2669,10 @@ static void parse_char_class(const char **regexp, struct re *re,
             to = parse_char(regexp, 0);
         }
     }
+    if (from > to) {
+        *error = REG_ERANGE;
+        goto error;
+    }
     add_re_char(re, from, to);
  error:
     return;
@@ -2839,6 +2961,7 @@ static char *re_cset_as_string(const struct re *re) {
 
     static const char *const empty_set = "[]";
     static const char *const total_set = "(.|\n)";
+    static const char *const not_newline = ".";
 
     char *result = NULL, *s;
     int from, to, negate;
@@ -2861,6 +2984,15 @@ static char *re_cset_as_string(const struct re *re) {
             /* Special case: the set matches every character */
             return strdup(total_set);
         }
+        if (from == '\n') {
+            for (from += 1;
+                 from <= UCHAR_MAX && bitset_get(re->cset, from);
+                 from += 1);
+            if (from > UCHAR_MAX) {
+                /* Special case: the set matches everything but '\n' */
+                return strdup(not_newline);
+            }
+        }
     }
 
     /* See if ']' and '-' will be explicitly included in the character set
@@ -3162,12 +3294,13 @@ static int re_collapse_trans(struct state *s1, struct state *s2,
  *     the transition INI -> FIN
  * (5) Convert that STRUCT RE to a string with RE_AS_STRING
  */
-int fa_as_regexp(struct fa *fa, char **regexp) {
+int fa_as_regexp(struct fa *fa, char **regexp, size_t *regexp_len) {
     int r;
     struct state *fin = NULL, *ini = NULL;
     struct re *eps = make_re(EPSILON);
 
     *regexp = NULL;
+    *regexp_len = 0;
     fa = fa_clone(fa);
 
     fin = add_state(fa,1);
@@ -3231,6 +3364,8 @@ int fa_as_regexp(struct fa *fa, char **regexp) {
         }
     }
     fa_free(fa);
+    if (*regexp != NULL)
+        *regexp_len = strlen(*regexp);
 
     return 0;
  error:
