@@ -357,12 +357,17 @@ void hash_set_allocator(hash_t *hash, hnode_alloc_t al,
 
 void hash_free_nodes(hash_t *hash)
 {
-    hscan_t hs;
-    hnode_t *node;
-    hash_scan_begin(&hs, hash);
-    while ((node = hash_scan_next(&hs))) {
-	hash_scan_delete(hash, node);
-	hash->freenode(node, hash->context);
+    hnode_t *node, *next;
+    hash_val_t chain;
+
+    for (chain = 0; chain < hash->nchains; chain ++) {
+      node = hash->table[chain];
+      while (node != NULL) {
+        next = node->next;
+        hash->freenode(node, hash->context);
+        node = next;
+      }
+      hash->table[chain] = NULL;
     }
     hash->nodecount = 0;
     clear_table(hash);
@@ -533,7 +538,7 @@ void hash_insert(hash_t *hash, hnode_t *node, const void *key)
     assert (hash_val_t_bit != 0);
     assert (node->next == NULL);
     assert (hash->nodecount < hash->maxcount);	/* 1 */
-    assert (hash_lookup(hash, key) == NULL);	/* 2 */
+    expensive_assert (hash_lookup(hash, key) == NULL);	/* 2 */
 
     if (hash->dynamic && hash->nodecount >= hash->highmark)	/* 3 */
 	grow_table(hash);
@@ -603,7 +608,7 @@ hnode_t *hash_delete(hash_t *hash, hnode_t *node)
     hash_val_t chain;
     hnode_t *hptr;
 
-    assert (hash_lookup(hash, node->key) == node);	/* 1 */
+    expensive_assert (hash_lookup(hash, node->key) == node);	/* 1 */
     assert (hash_val_t_bit != 0);
 
     if (hash->dynamic && hash->nodecount <= hash->lowmark
@@ -659,7 +664,7 @@ hnode_t *hash_scan_delete(hash_t *hash, hnode_t *node)
     hash_val_t chain;
     hnode_t *hptr;
 
-    assert (hash_lookup(hash, node->key) == node);
+    expensive_assert (hash_lookup(hash, node->key) == node);
     assert (hash_val_t_bit != 0);
 
     chain = node->hkey & hash->mask;
