@@ -263,10 +263,6 @@ char* read_file(const char *path);
  */
 const char *xstrerror(int errnum, char *buf, size_t len);
 
-/* A hidden flag used by augparse to suppress loading of all the modules
- * on the path */
-#define AUG_NO_MODL_AUTOLOAD (1 << 15)
-
 /* Struct: augeas
  * The data structure representing a connection to Augeas. */
 struct augeas {
