@@ -1,7 +1,7 @@
 /*
  * syntax.h: Data types to represent language syntax
  *
- * Copyright (C) 2007, 2008 Red Hat Inc.
+ * Copyright (C) 2007 - 2009 Red Hat Inc.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -210,12 +210,8 @@ struct regexp *regexp_make_empty(struct info *);
    regular expressions */
 void regexp_release(struct regexp *regexp);
 
-/* Construct a finite automaton from REGEXP. The pattern for REGEXP
- * must be known to be syntactically correct.
- *
- * Return a new automaton, or NULL if the REGEXP has a syntax error
- */
-fa_t regexp_to_fa(struct regexp *regexp);
+/* Produce a printable representation of R */
+char *regexp_escape(struct regexp *r);
 
 struct native {
     unsigned int argc;
@@ -223,51 +219,6 @@ struct native {
     struct value *(*impl)(void);
 };
 
-/*
- * Transformers for going from file globs to path names in the tree
- * functions are in transform.c
- */
-
-/* Filters for globbing files */
-struct filter {
-    unsigned int   ref;
-    struct filter *next;
-    struct string *glob;
-    unsigned int   include : 1;
-};
-
-struct filter *make_filter(struct string *glb, unsigned int include);
-void free_filter(struct filter *filter);
-
-/* Transformers that actually run lenses on contents of files */
-struct transform {
-    unsigned int      ref;
-    struct lens      *lens;
-    struct filter    *filter;
-};
-
-struct transform *make_transform(struct lens *lens, struct filter *filter);
-void free_transform(struct transform *xform);
-
-/* Load all files matching the TRANSFORM's filter into the tree in AUG by
- * applying the TRANSFORM's lens to their contents and putting the
- * resulting tree under "/files" + filename. Also stores some information
- * about filename underneath "/augeas/files" + filename
- */
-int transform_load(struct augeas *aug, struct transform *transform);
-
-/* Return 1 if TRANSFORM applies to PATH, 0 otherwise. The TRANSFORM
- * applies to PATH if (1) PATH starts with "/files/" and (2) the rest of
- * PATH matches the transform's filter
-*/
-int transform_applies(struct transform *transform, const char *path);
-
-/* Save TREE into the file corresponding to PATH. It is assumed that the
- * TRANSFORM applies to that PATH
- */
-int transform_save(struct augeas *aug, struct transform *transform,
-                   const char *path, struct tree *tree);
-
 /* An exception in the interpreter */
 struct exn {
     struct info *info;
@@ -369,6 +320,9 @@ struct term *make_app_term(struct term *func, struct term *arg,
                            struct info *info);
 struct term *make_app_ident(char *id, struct term *func, struct info *info);
 
+/* Duplicate a string; if STR is NULL, use the empty string "" */
+struct string *dup_string(const char *str);
+
 /* Make an EXN value
  * Receive ownership of INFO
  *
@@ -414,6 +368,8 @@ struct module *builtin_init(void);
 int __aug_load_module_file(struct augeas *aug, const char *filename);
 
 int interpreter_init(struct augeas *aug);
+
+struct lens *lens_lookup(struct augeas *aug, const char *qname);
 #endif
 
 
