@@ -32,7 +32,7 @@ run_tests ()
         fi
         printf "$action %-30s ... " $(basename $g .aug)
         set +e
-        errs=$(augparse -I ${MODULES} $g 2>&1 > /dev/null)
+        errs=$(augparse --nostdinc -I ${MODULES} $g 2>&1 > /dev/null)
         ret=$?
         set -e
         if [ $ret -eq $ret_fail ]; then
